"""Quiz data structures"""

from dataclasses import dataclass


@dataclass
class Source:
    url: str
    title: str | None = None
    description: str | None = None


@dataclass
class Image:
    url: str


@dataclass
class Answer:
    answer: str
    image: Image | None = None


@dataclass
class Choice:
    choice: str
    image: Image | None = None


@dataclass
class Question:
    question: str
    image: Image | None = None


@dataclass
class SingleChoiceQuestion:
    type = "single_choice"
    question: Question
    answer: Answer
    sources: list[Source]
    tags: list[str]
    notes: str | None = None
    hint: str | None = None


@dataclass
class MultipleChoiceQuestion:
    type = "multiple_choice"
    question: Question
    answer: Answer
    choices: list[Choice]
    sources: list[Source]
    tags: list[str]
    notes: str | None = None
    hint: str | None = None


@dataclass
class Round:
    title: str
    questions: list[MultipleChoiceQuestion | SingleChoiceQuestion]
    sub_title: str | None = None


@dataclass
class Quiz:
    title: str
    rounds: list[Round]
    sub_title: str | None = None
    slideshow: str | None = None



"""Reads quizes from files"""

from pathlib import Path
from typing import Optional
import string

import yaml
from pydantic import BaseModel, HttpUrl
from . import quiz as quiz_dataclasses


class Source(BaseModel):
    url: HttpUrl
    title: Optional[str] = None
    description: Optional[str] = None

    def to_dataclass(self) -> quiz_dataclasses.Source:
        return quiz_dataclasses.Source(
            description=self.description, url=str(self.url), title=self.title
        )


class Answer(BaseModel):
    answer: str
    image: Optional[str] = None

    def to_dataclass(self) -> quiz_dataclasses.Answer:
        return quiz_dataclasses.Answer(
            answer=self.answer,
            image=quiz_dataclasses.Image(url=self.image)
            if self.image is not None
            else None,
        )


class Question(BaseModel):
    question: str
    image: Optional[str] = None

    def to_dataclass(self) -> quiz_dataclasses.Question:
        return quiz_dataclasses.Question(
            question=self.question,
            image=quiz_dataclasses.Image(url=self.image)
            if self.image is not None
            else None,
        )


class SingleChoiceQuestion(BaseModel):
    question: str | Question
    answer: str | Answer
    source: Optional[str | Source] = None
    sources: Optional[list[str | Source]] = None
    tag: Optional[str] = None
    tags: Optional[list[str]] = None
    notes: str | None = None
    hint: str | None = None

    def to_dataclass(self) -> quiz_dataclasses.SingleChoiceQuestion:
        sources: list[str | Source] = [] if self.sources is None else self.sources
        if self.source:
            sources.append(self.source)
        tags = self.tags if self.tags is not None else []
        if self.tag:
            tags.append(self.tag)
        return quiz_dataclasses.SingleChoiceQuestion(
            answer=quiz_dataclasses.Answer(answer=self.answer)
            if isinstance(self.answer, str)
            else self.answer.to_dataclass(),
            question=quiz_dataclasses.Question(question=self.question)
            if isinstance(self.question, str)
            else self.question.to_dataclass(),
            sources=[
                quiz_dataclasses.Source(url=s)
                if isinstance(s, str)
                else s.to_dataclass()
                for s in sources
            ],
            tags=tags,
            notes=self.notes,
            hint=self.hint,
        )


class MultipleChoiceQuestion(BaseModel):
    question: str | Question
    choices: list[str | Answer]
    source: Optional[str | Source] = None
    sources: Optional[list[str | Source]] = None
    tag: Optional[str] = None
    tags: Optional[list[str]] = None
    notes: str | None = None
    hint: str | None = None

    def to_dataclass(self) -> quiz_dataclasses.MultipleChoiceQuestion:
        sources: list[str | Source] = [] if self.sources is None else self.sources
        if self.source:
            sources.append(self.source)
        tags = self.tags if self.tags is not None else []
        if self.tag:
            tags.append(self.tag)
        answers = [
            (i, a) for (i, a) in enumerate(self.choices) if isinstance(a, Answer)
        ]
        assert len(answers) == 1
        i, answer = answers[0]
        letter = string.ascii_lowercase[i]
        answer = answer.to_dataclass()
        answer.answer = f"{letter}. {answer.answer}"
        return quiz_dataclasses.MultipleChoiceQuestion(
            question=quiz_dataclasses.Question(question=self.question)
            if isinstance(self.question, str)
            else self.question.to_dataclass(),
            sources=[
                quiz_dataclasses.Source(url=s)
                if isinstance(s, str)
                else s.to_dataclass()
                for s in sources
            ],
            choices=[
                quiz_dataclasses.Choice(choice=a if isinstance(a, str) else a.answer)
                for a in self.choices
            ],
            answer=answer,
            tags=tags,
            notes=self.notes,
            hint=self.hint,
        )


class Round(BaseModel):
    title: str
    sub_title: Optional[str] = None
    questions: list[MultipleChoiceQuestion | SingleChoiceQuestion]

    def to_dataclass(self) -> quiz_dataclasses.Round:
        return quiz_dataclasses.Round(
            title=self.title,
            sub_title=self.sub_title,
            questions=[q.to_dataclass() for q in self.questions],
        )


class Quiz(BaseModel):
    title: str
    sub_title: Optional[str] = None
    rounds: list[Round]
    slideshow: Optional[str] = None

    def to_dataclass(self) -> quiz_dataclasses.Quiz:
        return quiz_dataclasses.Quiz(
            title=self.title,
            sub_title=self.sub_title,
            rounds=[r.to_dataclass() for r in self.rounds],
            slideshow=self.slideshow,
        )


def parse_file(filename: Path) -> quiz_dataclasses.Quiz:
    with filename.open("r") as fp:
        raw = yaml.safe_load(fp)
    quiz = Quiz.model_validate(raw)

    return quiz.to_dataclass()



import asyncio
import asyncio.exceptions
import collections
from pathlib import Path
import signal

import rich
import typer

from tablequiz import serialization, server

app = typer.Typer()


@app.command()
def parse(input: Path) -> None:
    quiz = serialization.parse_file(input)
    rich.print(quiz)


@app.command()
def serve(input: Path, host: str = "127.0.0.1", port: int = 5000) -> None:
    quiz = serialization.parse_file(input)
    srv = server.create_server(host=host, port=port, quiz=quiz)
    loop = asyncio.get_event_loop()

    # Try to exit a little more gracefully when run from watchfiles. Quietens
    # down one exception but there's still a asyncio.exceptions.CancelledError
    # from the lifespan loop. It's not important, but really bugs me :D. See
    # https://github.com/encode/uvicorn/issues/2173
    def signal_handler() -> None:
        srv.should_exit = True

    loop.add_signal_handler(signal.SIGINT, signal_handler)
    loop.add_signal_handler(signal.SIGTERM, signal_handler)

    asyncio.run(srv.serve())


def print_summary(totals: collections.Counter) -> None:
    rich.print(f"Questions: {totals['questions']}")
    rich.print(f"  Images: {totals['images']}")
    rich.print(f"  Multiple choice: {totals['multiple_choice']}")
    rich.print(f"  Single choice: {totals['single_choice']}")
    rich.print(
        f"  ratio single to multiple: {totals['single_choice'] / max(totals['multiple_choice'], 1)}"
    )
    tags = [
        (tag.split(":", 1)[1], count)
        for (tag, count) in totals.most_common()
        if tag.startswith("tag:")
    ]

    rich.print("Tags:")
    for tag, count in tags:
        rich.print(f"  {tag}: {count}")


@app.command()
def summary(input: Path) -> None:
    quiz = serialization.parse_file(input)

    rich.print(quiz.title)
    rich.print(quiz.sub_title)

    totals = collections.Counter()
    for i, round in enumerate(quiz.rounds):
        rich.print(f"\nRound {i+1}: {round.title}")
        rich.print(round.sub_title)
        round_totals = collections.Counter()
        for question in round.questions:
            round_totals["questions"] += 1
            round_totals[question.type] += 1
            round_totals["images"] += (
                1 if (question.question.image or question.answer.image) else 0
            )
            for tag in question.tags:
                round_totals[f"tag:{tag}"] += 1
            if not question.tags:
                round_totals["tag:no-tag"] += 1

        print_summary(round_totals)
        totals.update(round_totals)

    rich.print("\nQuiz totals:")

    print_summary(totals)


if __name__ == "__main__":
    typer.run(app)



import importlib.resources

import uvicorn
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Mount, Route
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates

from tablequiz.quiz import Quiz

with importlib.resources.path("tablequiz", "templates") as templates_path:
    templates = Jinja2Templates(directory=templates_path)


async def index(request: Request) -> Response:
    return templates.TemplateResponse(
        request, "index.html", context={"quiz": request.app.state.quiz}
    )


async def slideshow(request: Request) -> Response:
    return templates.TemplateResponse(
        request, "slideshow.html", context={"quiz": request.app.state.quiz}
    )


async def quiz_round(request: Request) -> Response:
    round_number = request.path_params["round"]
    quiz = request.app.state.quiz
    quiz_round = quiz.rounds[round_number - 1]
    return templates.TemplateResponse(
        request,
        "quiz_round.html",
        context={"quiz": quiz, "round": quiz_round},
    )


async def quiz_round_answers(request: Request) -> Response:
    round_number = request.path_params["round"]
    quiz = request.app.state.quiz
    quiz_round = quiz.rounds[round_number - 1]
    return templates.TemplateResponse(
        request,
        "quiz_answers.html",
        context={"quiz": quiz, "round": quiz_round},
    )


routes: list[Route | Mount] = [
    Route("/", endpoint=index, name="index"),
    Route("/slideshow", endpoint=slideshow, name="slideshow"),
    Route(
        "/rounds/{round:int}/answers", endpoint=quiz_round_answers, name="round_answers"
    ),
    Route("/rounds/{round:int}", endpoint=quiz_round, name="round"),
    Mount("/static", StaticFiles(packages=["tablequiz"]), name="static"),
    Mount("/images", StaticFiles(directory="images"), name="images"),
]


def create_server(quiz: Quiz, host: str, port: int) -> uvicorn.Server:
    app = Starlette(debug=True, routes=routes)
    app.state.quiz = quiz
    config = uvicorn.Config(
        app=app,
        host=host,
        port=port,
        log_level="info",
    )
    return uvicorn.Server(config)
